
#include "stdafx.h"

#include "ManagedLock.h"
#include "CreateBonInstance.h"
#include "DataQueue.h"
#include "DataQueueManager.h"
#include "TuningSpaceInformation.h"
#include "EpgData.h"
#include "ProgramInformation.h"
#include "Tuner.h"

namespace Bon
{
	Tuner::Tuner(ITunerDefinition^ tunerDefinition)
		: tunerDefinition_(tunerDefinition), running_(false), enableEmmProcess_(false), enableLnbPower_(true), tuningSpaces_(nullptr),
		pTsSourceTuner_(0), pTsPacketSync_(0), pProgAnalyzer_(0), pTsDescrambler_(0), pManager_(0)
	{
		pTsSourceTuner_ = Internal::CreateBonInstance<ITsSourceTuner>(TEXT("CTsSourceTuner"));
		pTsPacketSync_ = Internal::CreateBonInstance<ITsPacketSync>(TEXT("CTsPacketSync"));
		pProgAnalyzer_ = Internal::CreateBonInstance<IProgAnalyzer>(TEXT("CProgAnalyzer"));
		pTsDescrambler_ = Internal::CreateBonInstance<ITsDescrambler>(TEXT("CTsDescrambler"));
		pManager_ = new Internal::DataQueueManager();

		dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pTsPacketSync_));
		dynamic_cast<IMediaDecoder *>(pTsPacketSync_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pTsDescrambler_));
		dynamic_cast<IMediaDecoder *>(pTsDescrambler_)->SetOutputDecoder(dynamic_cast<IMediaDecoder *>(pProgAnalyzer_));
		dynamic_cast<IMediaDecoder *>(pProgAnalyzer_)->SetOutputDecoder(pManager_);

		pin_ptr<const wchar_t> pTunerName = PtrToStringChars(tunerDefinition_->TunerName);

		int tunerResult = pTsSourceTuner_->OpenTuner(pTunerName);
		if (tunerResult > 0)
		{
			throw gcnew BonException("Cannot open tuner: Name={0}, Result={1}",
				tunerDefinition_->TunerName, tunerResult);
		}

		if (pTsDescrambler_->OpenDescrambler(TEXT("CBcasCardReader")) == false)
		{
			pTsSourceTuner_->CloseTuner();
			throw gcnew BonException("Cannot open card reader");
		}

		pTsPacketSync_->DiscardNullPacket(true);
		pTsDescrambler_->DiscardScramblePacket(false);
		pTsDescrambler_->EnableBuffering(true);
	}

	Tuner::~Tuner()
	{
		Internal::ManagedLock lock(this);

		BON_SAFE_RELEASE(pTsSourceTuner_);
		BON_SAFE_RELEASE(pTsPacketSync_);
		BON_SAFE_RELEASE(pTsDescrambler_);
		BON_SAFE_RELEASE(pProgAnalyzer_);
		BON_SAFE_DELETE(pManager_);
	}

	System::String^ Tuner::TunerName::get()
	{
		return tunerDefinition_->TunerName;
	}

	System::String^ Tuner::DeviceName::get()
	{
		return tunerDefinition_->DeviceName;
	}

	array<ITuningSpaceInformation^>^ Tuner::TuningSpaces::get()
	{
		return tunerDefinition_->TuningSpaces;
	}

	int Tuner::DeviceID::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetDeviceID();
	}

	bool Tuner::IsRunning::get()
	{
		return running_;
	}

	array<System::Byte>^ Tuner::BcasCardID::get()
	{
		Internal::ManagedLock lock(this);

		CMediaData mediaData(0);
		pTsDescrambler_->GetBcasCardID(&mediaData);

		array<System::Byte>^ buffer = gcnew array<System::Byte>(mediaData.GetSize());
		pin_ptr<BYTE> pBuffer = &buffer[0];
		memcpy(pBuffer, mediaData.GetData(), mediaData.GetSize());

		return buffer;
	}

	bool Tuner::EnableEmmProcess::get()
	{
		return enableEmmProcess_;
	}

	void Tuner::EnableEmmProcess::set(bool value)
	{
		Internal::ManagedLock lock(this);

		if (enableEmmProcess_ != value)
		{
			enableEmmProcess_ = value;
			if (running_ == true)
			{
				pTsDescrambler_->EnableEmmProcess(enableEmmProcess_);
			}
		}
	}

	bool Tuner::EnableLnbPower::get()
	{
		return enableLnbPower_;
	}

	void Tuner::EnableLnbPower::set(bool value)
	{
		Internal::ManagedLock lock(this);

		if (enableLnbPower_ != value)
		{
			enableLnbPower_ = value;
			if (running_ == true)
			{
				pTsSourceTuner_->SetLnbPower(enableLnbPower_);
			}
		}
	}

	float Tuner::SignalLevel::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetSignalLevel();
	}

	int Tuner::TuningSpace::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetCurSpace();
	}

	void Tuner::TuningSpace::set(int space)
	{
		Internal::ManagedLock lock(this);

		pTsSourceTuner_->SetChannel(space, 0);
	}

	int Tuner::Channel::get()
	{
		Internal::ManagedLock lock(this);

		return pTsSourceTuner_->GetCurChannel();
	}

	void Tuner::Channel::set(int channel)
	{
		Internal::ManagedLock lock(this);

		pTsSourceTuner_->SetChannel(pTsSourceTuner_->GetCurSpace(), channel);
	}

	int Tuner::NetworkID::get()
	{
		Internal::ManagedLock lock(this);

		return pProgAnalyzer_->GetNetworkID();
	}

	System::String^ Tuner::NetworkName::get()
	{
		Internal::ManagedLock lock(this);

		const CBonString networkName = pProgAnalyzer_->GetNetworkName();
		if (networkName.IsEmpty())
		{
			throw gcnew BonException("Cannot get network name");
		}

		return gcnew System::String(networkName);
	}

	int Tuner::TransportStreamID::get()
	{
		Internal::ManagedLock lock(this);

		return static_cast<int>(pProgAnalyzer_->GetTsID());
	}

	System::DateTime Tuner::StreamTime::get()
	{
		Internal::ManagedLock lock(this);

		SYSTEMTIME streamTime;
		memset(&streamTime, 0, sizeof streamTime);
		bool result = pProgAnalyzer_->GetStreamTime(&streamTime);
		if (result == false)
		{
			throw gcnew BonException("Cannot get stream time");
		}

		// GetStreamTime() is local time
		return System::DateTime(streamTime.wYear, streamTime.wMonth, streamTime.wDay,
			streamTime.wHour, streamTime.wMinute, streamTime.wSecond, streamTime.wMilliseconds,
			System::DateTimeKind::Local);
	}

	array<IProgramInformation^>^ Tuner::Programs::get()
	{
		System::Collections::Generic::List<IProgramInformation^>^ programs =
			gcnew System::Collections::Generic::List<IProgramInformation^>();

		Internal::ManagedLock lock(this);

		for (WORD i = 0; i < pProgAnalyzer_->GetProgramNum(); i++)
		{
			WORD programID = pProgAnalyzer_->GetProgramID(i);

			WORD programVideoID = pProgAnalyzer_->GetProgramVideoPID(programID);
			WORD programAudioID = pProgAnalyzer_->GetProgramAudioPID(programID);
			WORD programPcrID = pProgAnalyzer_->GetProgramPcrPID(programID);

			const CBonString serviceName = pProgAnalyzer_->GetServiceName(programID);

			const IEpgItem *pEpgItem = pProgAnalyzer_->GetCurrentEpgInfo(programID);

			programs->Add(gcnew ProgramInformation(
				programID, programVideoID, programAudioID, programPcrID,
				gcnew System::String(serviceName), gcnew EpgData(pEpgItem)));
		}

		return programs->ToArray();
	}

	int Tuner::ProgramID::get()
	{
		Internal::ManagedLock lock(this);

		return pProgAnalyzer_->GetCurrentProgramID();
	}

	void Tuner::ProgramID::set(int value)
	{
		Internal::ManagedLock lock(this);

		if (pProgAnalyzer_->SetCurrentProgramID(static_cast<WORD>(value)) == false)
		{
			throw gcnew BonException("Cannot set program ID");
		}
	}

	DescramblingStates Tuner::DescramblingStatus::get()
	{
		Internal::ManagedLock lock(this);

		return static_cast<DescramblingStates>(
			pTsDescrambler_->GetDescramblingState(pProgAnalyzer_->GetCurrentProgramID()));
	}

	IDataQueue^ Tuner::CreateQueue(__int64 maxSize)
	{
		Internal::ManagedLock lock(this);

		System::Collections::Generic::KeyValuePair<IDataQueue^, int> queue =
			pManager_->CreateQueue(this, maxSize);
		if (queue.Value == 1)
		{
			pTsDescrambler_->EnableEmmProcess(enableEmmProcess_);
			pTsSourceTuner_->SetLnbPower(enableLnbPower_);
			dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->PlayDecoder();

			running_ = true;
		}
		return queue.Key;
	}

	void Tuner::RemoveQueue(DataQueue^ queue)
	{
		Internal::ManagedLock lock(this);

		int count = pManager_->RemoveQueue(queue);
		if (count == 0)
		{
			if (pTsSourceTuner_ != nullptr)
			{
				dynamic_cast<IMediaDecoder *>(pTsSourceTuner_)->StopDecoder();
				pTsDescrambler_->EnableEmmProcess(false);
				pTsSourceTuner_->SetLnbPower(false);

				running_ = false;
			}
		}
	}
}
